#include "sq.h"

#include "FilePlugin.h"			/* sqFileInit() */
#include "JoystickTabletPlugin.h"	/* joystickInit() */

#define USE_XMODEM
#ifdef USE_XMODEM
#include "xmodem.h"
int using_xmodem = 1;
#endif


/* OS_TYPE may be set in configure.in and passed via the Makefile */

#ifndef OS_TYPE
# ifdef UNIX
#   define OS_TYPE "unix"
# else
#  define OS_TYPE "unknown"
# endif
#endif

#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>

#include <fcntl.h>
#include <sys/mman.h>

#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif


#if defined(HAVE_LIBDL)
# define USE_INTERNAL_IMAGE
# include <dlfcn.h>
# if !defined(NDEBUG)
#   define NDEBUG
# endif
# include <assert.h>
# include "zipio.h"
#endif

#define TX39_TPANEL
#define TX39_BUTTONS

/*** Variables -- Imported from Virtual Machine ***/
extern unsigned char *memory;
extern int interruptKeycode;
extern int interruptPending;
extern int interruptCheckCounter;
extern int savedWindowSize;

/*** Variables -- image and path names ***/
#define IMAGE_NAME_SIZE MAXPATHLEN

static char shortImageName[MAXPATHLEN+1];	/* image name */
char        imageName[MAXPATHLEN+1];		/* full path to image */
static char vmPath[MAXPATHLEN+1];		/* full path to image directory */
static char vmName[MAXPATHLEN+1];		/* full path to vm */

#define DefaultHeapSize		20	/* megabytes */

int initialHeapSize= DefaultHeapSize * 1024 * 1024;

/*** Variables -- globals for access from pluggable primitives ***/
int    argCnt= 0;
char **argVec= 0;
char **envVec= 0;

/* Pointers to arguments for VM.  Used by getAttributeIntoLength(). */
int    vmArgCnt= 0;
char **vmArgVec= 0;

/* Pointers to arguments for Squeak.  Used by getAttributeIntoLength(). */
int    squeakArgCnt= 0;
char **squeakArgVec= 0;

#if defined(USE_INTERNAL_IMAGE)
unsigned char *internalImage=    0;	/* non-zero means use internal image */
unsigned char *internalImagePtr= 0;	/* non-zero means plain image */
z_stream      *internalZStream=  0;	/* non-zero means gzipped image */
#endif

#undef USE_ITIMER			/* severely confuses GNU's profil() */

#ifdef USE_ITIMER
unsigned int	lowResMSecs= 0;
#define	LOW_RES_TICK_MSECS	20	/* 1/50 second resolution */
#endif

#define SqueakWhite	0
#define SqueakBlack	1

#define TOO_LAZY_TO_REBUILD_SOURCES
#ifdef TOO_LAZY_TO_REBUILD_SOURCES
void *browserWindow, *stDisplay, *stWindow;

void XSendEvent() { }
void XInternAtom() { }
/* void XStoreName() { } */
#endif

int		 asmAlign= 1;
int		 sleepWhenUnmapped= 0;
int		 noJitter= 0;
int		 withSpy= 0;
int		 noTitle= 0;
int		 fullScreen= 0;
struct timeval	 startUpTime;

/* maximum input polling frequency */
#define	MAXPOLLSPERSEC	33

/* polling functions in socket/sound plugins, called from HandleEvents */

void (*socketPollFunction)(int delay, int extraFd)= 0;	/* aioPollForIO */
void (*soundPollFunction)(void)= 0;			/* auPollForIO */


/* fb stuff */

/*** Variables -- Event Recording ***/
#define KEYBUF_SIZE 64

int keyBuf[KEYBUF_SIZE];	/* circular buffer */
int keyBufGet= 0;		/* index of next item of keyBuf to read */
int keyBufPut= 0;		/* index of next item of keyBuf to write */
int keyBufOverflows= 0;		/* number of characters dropped */

int buttonState= 0;		/* mouse button and modifier state when mouse
				   button went down or 0 if not pressed */

struct mypoint {
  int x;
  int y;
};

struct mypoint mousePosition;


int              secure= 0;		/* true if running as a browser plugin */
int		 headless= 0;

/*** Functions ***/

#ifdef ioMSecs
# undef ioMSecs
#endif

int  HandleEvents(void);
void RecordFullPathForImageName(char *localImageName);
void SetUpTimers(void);
void usage(void);
void imageNotFound(char *imageName);
void ParseArguments(int argc, char **argv);
void segv(int ignored);

time_t convertToSqueakTime(time_t);	/* unix epoch -> Squeak epoch */

#include <linux/fb.h>
#include <linux/vt.h>
#include <linux/kd.h>

#define FB_DEV "/dev/fb0"

static int fb_fh;

static struct fb_fix_screeninfo fb_finfo;
static struct fb_var_screeninfo fb_vinfo;

static char *fbmem;
static int fbsize;

int openFbDisplay()
{
  int bits;
  if ((fb_fh = open(FB_DEV, O_RDWR)) < 0) {
    perror("open" FB_DEV);
    return -1;
  }

  if (ioctl(fb_fh, FBIOGET_FSCREENINFO, &fb_finfo)) {
    perror("ioctl" FB_DEV);
    return -1;
  }

  ioctl(fb_fh, FBIOGET_VSCREENINFO, &fb_vinfo);

  bits = fb_vinfo.bits_per_pixel;
  fbsize = (fb_vinfo.xres * bits >> 3) * fb_vinfo.yres;

  fbmem = mmap(0, fbsize, PROT_READ | PROT_WRITE, MAP_SHARED, fb_fh, 0);
  if (fbmem < 0) {
    perror("mmap() " FB_DEV);
    return -1;
  }

      /* horrible hack */
    { 
            int xx = open("/dev/tty1", O_RDWR);
            if (xx < 0) {
                    perror("open for icky hack");
                    return -1;
            }
            ioctl(xx, KDSETMODE, KD_GRAPHICS);
            close(xx);
    }


  /* put some random garbage on screen, just to demonstrate we're in
     graphics mode now */
  memcpy(fbmem, openFbDisplay, fbsize);

  return 0;
}

int closeFbDisplay(void) 
{
  close(fb_fh);
  return 0;
}

/* tpanel */

#ifdef TX39_TPANEL

#define TPANEL_DEV "/dev/tpanel"
static int tp_fd;

typedef unsigned short tpanel_sample;

#define TRANSFORMATION_UNITS_PER_PIXEL 4

typedef struct
{
	/*
	 * Coefficients for the transformation formulas:
	 *
	 *     m = (ax + by + c) / s
	 *     n = (dx + ey + f) / s
	 *
	 * These formulas will transform a device point (x, y) to a
	 * screen point (m, n) in fractional pixels.  The fraction
	 * is 1 / TRANSFORMATION_UNITS_PER_PIXEL.
	 */

	int a, b, c, d, e, f, s;
} TRANSFORMATION_COEFFICIENTS;

typedef struct
{
        int x, y;
} XYPOINT;


int enable_pointing_coordinate_transform = 1;

static TRANSFORMATION_COEFFICIENTS tc;

int GetPointerCalibrationData()
{
        extern int errno;
	/*
	 * Read the calibration data from the calibration file.
	 * Calibration file format is seven coefficients separated by spaces.
	 */

	/* Get pointer calibration data from this file */
	const char cal_filename[] = "/etc/pointercal";

	int items;

	FILE* f = fopen(cal_filename, "r");
	if ( f == NULL )
	{
		fprintf(stderr, "Error %d opening pointer calibration file %s.\n",
			errno, cal_filename);
		return -1;
	}

	items = fscanf(f, "%d %d %d %d %d %d %d",
		&tc.a, &tc.b, &tc.c, &tc.d, &tc.e, &tc.f, &tc.s);
	if ( items != 7 )
	{
		fprintf(stderr, "Improperly formatted pointer calibration file %s.\n",
			cal_filename);
		return -1;
	}

#ifdef TEST
		printf("a=%d b=%d c=%d d=%d e=%d f=%d s=%d\n",
			tc.a, tc.b, tc.c, tc.d, tc.e, tc.f, tc.s);
#endif

	return 0;
}

inline XYPOINT DeviceToScreen(XYPOINT p)
{
	/*
	 * Transform device coordinates to screen coordinates.
	 * Take a point p in device coordinates and return the corresponding
	 * point in screen coodinates.
	 * This can scale, translate, rotate and/or skew, based on the coefficients
	 * calculated above based on the list of screen vs. device coordinates.
	 */

	static XYPOINT prev;
	/* set slop at 3/4 pixel */
	const short slop = TRANSFORMATION_UNITS_PER_PIXEL * 3 / 4;
	XYPOINT new, out;

	/* transform */
	new.x = (tc.a * p.x + tc.b * p.y + tc.c) / tc.s;
	new.y = (tc.d * p.x + tc.e * p.y + tc.f) / tc.s;

	/* hysteresis (thanks to John Siau) */
	if ( abs(new.x - prev.x) >= slop )
		out.x = (new.x | 0x3) ^ 0x3;
	else
		out.x = prev.x;

	if ( abs(new.y - prev.y) >= slop )
		out.y = (new.y | 0x3) ^ 0x3;
	else
		out.y = prev.y;

	prev = out;

	return out;
}

int openTpanel(void)
{
  if ((tp_fd = open (TPANEL_DEV, O_RDONLY | O_NDELAY)) < 0) {
    perror("open " TPANEL_DEV);
          return -1;
    }

    if (GetPointerCalibrationData() < 0) {
          fprintf(stderr, "open of calibration data failed.\n");
          return -1;
  }
}

#endif

#ifdef TX39_BUTTONS

#define BUTTONS_DEV "/dev/buttons"

static int bt_fd;

static int button_two = 0;
static int button_three = 0;

int openButtons(void)
{
  if ((bt_fd = open (BUTTONS_DEV, O_RDONLY | O_NDELAY)) < 0) {
    perror("open " BUTTONS_DEV);
          return -1;
    }
  return 0;
}

#endif

/*** VM Home Directory Path ***/

int vmPathSize(void)
{
  return strlen(vmPath);
}

int vmPathGetLength(int sqVMPathIndex, int length)
{
  char *stVMPath= (char *)sqVMPathIndex;
  int count, i;

  count= strlen(vmPath);
  count= (length < count) ? length : count;

  /* copy the file name into the Squeak string */
  for (i= 0; i < count; i++)
    stVMPath[i]= vmPath[i];

  return count;
}


/*** I/O Primitives ***/


int ioFormPrint(int bitsAddr, int width, int height, int depth,
		double hScale, double vScale, int landscapeFlag)
{
  fprintf(stderr, "no printing from framebuffer\n");
  return false;
}

int ioBeep(void)
{
  return 0;
}

int ioGetButtonState(void)
{
  ioProcessEvents();  /* process all pending events */
#ifndef HEADLESS
  return buttonState;
#else
  return 0;
#endif
}

int ioGetKeystroke(void)
{
  int keystate;

  ioProcessEvents();  /* process all pending events */
#ifdef HEADLESS
  return -1;  /* keystroke buffer is empty */
#else

  if (keyBufGet == keyBufPut)
    return -1;  /* keystroke buffer is empty */

  keystate= keyBuf[keyBufGet];
  keyBufGet= (keyBufGet + 1) % KEYBUF_SIZE;
  /* set modifer bits in buttonState to reflect the last keystroke fetched */
#if 0 /* FIXME, not recording key up events properly */
  buttonState= ((keystate >> 5) & 0xF8) | (buttonState & 0x7);
#endif

  return keystate;

  return -1;
#endif
}

int ioLowResMSecs(void)
{
#ifdef USE_ITIMER
  return lowResMSecs;
#else
  return ioMSecs();
#endif
}

int ioMSecs(void)
{
  struct timeval now;
  gettimeofday(&now, 0);
  if ((now.tv_usec-= startUpTime.tv_usec) < 0) {
    now.tv_usec+= 1000000;
    now.tv_sec-= 1;
  }
  now.tv_sec-= startUpTime.tv_sec;
  return (now.tv_usec / 1000 + now.tv_sec * 1000);
}

int ioMicroMSecs(void)
{
  /* return the highest available resolution of the millisecond clock */
  return ioMSecs();	/* this already to the nearest millisecond */
}

int ioRelinquishProcessorForMicroseconds(int microSeconds)
{
  char oldScreenByte;

  oldScreenByte = fbmem[0];

  fbmem[0] = 255;

  /* sleep in select() for immediate response to socket i/o */
  if (socketPollFunction != 0)
#if 0    
#ifdef HEADLESS
    socketPollFunction(microSeconds, 0);
#else
    socketPollFunction(microSeconds, stXfd);
#endif
#else

#ifdef TX39_TPANEL
    socketPollFunction(microSeconds, tp_fd);
#else
    socketPollFunction(microSeconds, 0); /* TODO */
#endif

#endif
  else
    {
      /* can't rely on BSD usleep */
      struct timeval tv;
      tv.tv_sec=  microSeconds / 1000000;
      tv.tv_usec= microSeconds % 1000000;
      select(0, 0, 0, 0, &tv);
    }

  fbmem[0] = oldScreenByte;

  return microSeconds;
}



int ioMousePoint(void)
{
  ioProcessEvents();  /* process all pending events */
  /* x is high 16 bits; y is low 16 bits */
#ifndef HEADLESS
  return (mousePosition.x << 16) | (mousePosition.y);
#else
  return 0;
#endif
}

int ioPeekKeystroke(void)
{
#ifdef HEADLESS
  ioProcessEvents();  /* process all pending events */
  return -1;  /* keystroke buffer is empty */
#else

  int keystate;

  ioProcessEvents();  /* process all pending events */

  if (keyBufGet == keyBufPut)
    return -1;  /* keystroke buffer is empty */

  keystate= keyBuf[keyBufGet];
  /* set modifer bits in buttonState to reflect the last keystroke peeked at */
#if 0 /* FIXME */
  buttonState= ((keystate >> 5) & 0xF8) | (buttonState & 0x7);
#endif
  return keystate;

#if 0
  return -1;
#endif

#endif
}



int ioProcessEvents(void) 
{
  static unsigned long nextPollTick= 0;

  if ((unsigned long)ioLowResMSecs() > nextPollTick)
    {
      /* time to process events! */
      while (HandleEvents())
	{
	  /* process all pending events */
	}
    /* wait a while before trying again */
      nextPollTick= ioLowResMSecs() + (1000 / MAXPOLLSPERSEC);
    }
  return 0;
}

void keyBufAppend(int keystate)
{
  /* bug: this should be rewritten to cope with nConv > 1 */
  keyBuf[keyBufPut]= keystate;
  keyBufPut= (keyBufPut + 1) % KEYBUF_SIZE;
  if (keyBufGet == keyBufPut)
    {
      /* buffer overflow; drop the last character */
      keyBufGet= (keyBufGet + 1) % KEYBUF_SIZE;
      keyBufOverflows++;
    }
}


#ifdef TX39_TPANEL
static int handleTpanel(void) 
{
  unsigned short tp_evt[3];
  XYPOINT transformed;
  int count;

  count = read(tp_fd, tp_evt, sizeof(tp_evt));
  if (count < sizeof(tp_evt)) {
    return 0;
  }
  
  switch (tp_evt[0]) {
  case 0:
    /* pen release event.  ignore x/y, they're probably bogus */
    buttonState = 0;
    /* fprintf(stderr, "release\n"); */
    return 1;
    break;
  case 1:
  case 2:
    if (tp_evt[1] < 10 || tp_evt[2] < 10)
      return 0;
#ifdef TX39_BUTTONS
    if (button_three) {
      buttonState = 1;
    } else if (button_two) {
      buttonState = 2;
    } else {
      buttonState = 4;
    }
#else    
    buttonState = 4;
#endif

    transformed.x = tp_evt[1];
    transformed.y = tp_evt[2];
    transformed = DeviceToScreen(transformed);
    mousePosition.x = transformed.x >>2;
    mousePosition.y = transformed.y >>2;
    /* fprintf(stderr, "mouse %d %d\n", mousePosition.x, mousePosition.y); */
    return 1;
  }
}
#endif

#ifdef TX39_BUTTONS
static int handleButtons(void)
{
  int count;
  char c, code;
  int press;

  count = read(bt_fd, &c, 1);
  if (count < 1) {
    return 0;
  }

  /* fprintf(stderr, "got button %x\n", c); */

  code = c & 0x7f;
  press = (c & 0x80) ? 0 : 1;

  switch (code) {
  case 'S':
    if (press) {
      interruptPending = true;
      interruptCheckCounter = 0;
    }
    break;
  case 'U':
    if (press) {
      keyBufAppend(30);
    }
    break;
  case 'D':
    if (press) {
      keyBufAppend(31);
    }
    break;
  case '2':
    if (press) {
      keyBufAppend('r' | (1 <<11)); /* command r, recognizer */
    }
    break;
  case '1':
    if (press) {
      /* keyBufAppend('O' | (1<<11) | (1<<8)); */ /* command shift O */
      keyBufAppend('\r');
    }
    break;
  case 'P':
    button_two = press;
    break;
  case '3':
    button_three = press;
    break;
  default:
    /* do nothing */
  }

  return 1;
}
#endif    
  
  


#if defined(TX39_TPANEL) || defined(TX39_BUTTONS)
#define HANDLE_EVENTS_SELECT
#endif

int HandleEvents(void) 
{
  /* TODO */
#ifdef HANDLE_EVENTS_SELECT
  int count;

  fd_set fds;
  struct timeval tv = {0,0};
  int maxfd = 0;

#endif


  /* quick checks for asynchronous socket/sound i/o */
  if (socketPollFunction != 0)
    socketPollFunction(0, 0);
  if (soundPollFunction != 0)
    soundPollFunction();

#ifdef HANDLE_EVENTS_SELECT
  
  FD_ZERO(&fds);

#ifdef TX39_TPANEL
  FD_SET(tp_fd, &fds);
  maxfd = tp_fd;
#endif

#ifdef TX39_BUTTONS
  FD_SET(bt_fd, &fds);
  if (bt_fd > maxfd) {
    maxfd = bt_fd;
  }
#endif

  count = select(maxfd + 1, &fds, NULL, NULL, &tv);

  if (count <= 0)
    return 0;

  /* fprintf(stderr, "made it through select\n"); */

#ifdef TX39_BUTTONS
  if (FD_ISSET(bt_fd, &fds)) {
    if (handleButtons())
      return 1;
  }
#endif

#ifdef TX39_TPANEL
  if (FD_ISSET(tp_fd, &fds)) {
    if (handleTpanel())
      return 1;
  }
#endif

#endif /* HANDLE_EVENTS_SELECT */

  return 0;
}



/*** Command line ***/



static char *progName= 0;

void usage()
{
  printf("Usage: %s [<option>...] [<imageName> [<argument>...]]\n", progName);
  printf("       %s [<option>...] -- [<argument>...]\n\n", progName);
  printf("<option> can be:\n");
  printf("  -align <n>           align functions at n-bytes (jit)\n");
  printf("  -display <dpy>       display on <dpy> (default: $DISPLAY)\n");
  printf("  -fullscreen          occupy the entire screen\n");
  printf("  -headless            run in headless (no window) mode\n");
  printf("  -help                print this help message, then exit\n");
  printf("  -lazy                go to sleep when main window unmapped\n");
  printf("  -memory <size>[mk]   set initial memory size (default: %dm)\n",
	 DefaultHeapSize);
  printf("  -nojit               disable the dynamic (runtime) compiler\n");
  printf("  -notitle             disable the Squeak window title bar\n");
  printf("  -secure              enable file sandbox\n");
  printf("  -spy                 enable the system spy (jit)\n");
  printf("  -version             print version information, then exit\n");
  printf("  -xasync              don't serialize display updates\n");
  printf("  -xshm                use X shared memory extension\n");
  printf("\nNotes:\n");
  printf("  <imageName> defaults to `Squeak"SQ_VERSION".image'.\n");
  printf("  Using `unix:0' for <dpy> may improve local display performance.\n");
  printf("  -xshm only works when Squeak is running on the X server host.\n");
  printf("  <argument>s are ignored, but are processed by the Squeak image.\n");
  printf("  The first <argument> normally names a Squeak `script' to execute.\n");
  printf("  Precede <arguments> by `--' to use default image.\n");
  exit(1);
}


void imageNotFound(char *imageName)
{
  /* image file is not found */
  fprintf(stderr,"\
Could not open the Squeak image file `%s'.

There are three ways to open a Squeak image file.  You can:
  1. Put copies of the default image and changes files in this directory.
  2. Put the name of the image file on the command line when you
     run squeak (use the `-help' option for more information).
  3. Set the environment variable SQUEAK_IMAGE to the name of the image
     that you want to use by default.\n", imageName);
  exit(1);
}


void versionInfo(void)
{
  extern int vm_serial;
  extern char *vm_date, *cc_version, *ux_version;

  fprintf(stderr, "%s %s #%d %s%s %s\n%s\n",
	  VM_HOST, SQ_VERSION, vm_serial,
# if defined(USE_XSHM)
	  "XShm ",
# else
	  " ",
# endif
	  vm_date, cc_version, ux_version);
  fprintf(stderr, "default plugin location: %s/%s*.so\n",
	  SQ_LIBDIR, SQ_MODULE_PREFIX);
  {
    static char *ipl[]= {
#   include "sqUnixInternalPlugins.h"
      (char *)0 };
    char **iplp= 0;
    fprintf(stderr, "internal plugins:\n");
    for (iplp= ipl; *iplp != (char *)0; ++iplp)
      fprintf(stderr, "  %s\n", *iplp);
  }
  exit(0);
}


void outOfMemory(void)
{
  fprintf(stderr, "out of memory\n");
  exit(1);
}


int strtobkm(char *str)
{
  char *suffix;
  int value= strtol(str, &suffix, 10);
  switch (*suffix)
    {
    case 'k': case 'K':
      value*= 1024;
      break;
    case 'm': case 'M':
      value*= 1024*1024;
      break;
    }
  return value;
}


void ParseEnvironment(void)
{
  char *ev= 0;

  if ((ev= getenv("SQUEAK_ALIGN")))	asmAlign= atoi(ev);
  if (asmAlign <  1) asmAlign= 1;
  if (asmAlign > 16) asmAlign= 16;

  if ((ev= getenv("SQUEAK_IMAGE")))	strcpy(shortImageName, ev);
  else					strcpy(shortImageName,
					       "Squeak"SQ_VERSION".image");
  if ((ev= getenv("SQUEAK_MEMORY")))	initialHeapSize= strtobkm(ev);
  if (getenv("SQUEAK_LAZY"))		sleepWhenUnmapped= 1;
  if (getenv("SQUEAK_NOJIT"))		noJitter= 1;
  if (getenv("SQUEAK_SPY"))		withSpy= 1;
  if (getenv("SQUEAK_NOTITLE"))		noTitle= 1;
  if (getenv("SQUEAK_FULLSCREEN"))	fullScreen= 1;
  if (getenv("SQUEAK_SECURE"))		secure= 1;
}


void ParseArguments(int argc, char *argv[])
{
# define skipArg()	(--argc, argv++)
# define saveArg()	(vmArgVec[vmArgCnt++]= *skipArg())

  saveArg();		/* vm name */

  while ((argc > 0) && (**argv == '-'))	/* more options to parse */
    {
      char *arg= *argv;
      if (!strcmp(arg, "--"))		/* escape from option processing */
	break;

      saveArg();

      if      (!strcmp(arg, "-help"))		usage();
      else if (!strcmp(arg, "-headless"))	headless= 1;
      else if (!strcmp(arg, "-lazy"))		sleepWhenUnmapped= 1;
      else if (!strcmp(arg, "-notitle"))	noTitle= 1;
      else if (!strcmp(arg, "-nojit"))		noJitter= 1;
      else if (!strcmp(arg, "-spy"))		withSpy= 1;
      else if (!strcmp(arg, "-fullscreen"))	fullScreen= 1;
      else if (!strcmp(arg, "-secure"))		secure= true;
      else if (!strcmp(arg, "-version"))	versionInfo();
      else
	{
	  /* option requires an argument */
	  if      (arg == 0)			usage();
	  else if (!strcmp(arg, "-align"))
	    {
	      asmAlign= atoi(saveArg());
	      if (asmAlign <  1) asmAlign= 1;
	      if (asmAlign > 16) asmAlign= 16;
	    }
	  else if (!strcmp(arg, "-memory"))	initialHeapSize= strtobkm(saveArg());
	  else
	    usage();
	} /* option with argument */
    } /* while(more options) */

  if (argc > 0)
    {
      if (!strcmp(*argv, "--"))
	skipArg();
      else
	{
	  strcpy(shortImageName, saveArg());
	  if (0 == strstr(shortImageName, ".image"))
	    strcat(shortImageName, ".image");
	}
    }
  /* save remaining arguments as Squeak arguments */
  while (argc > 0)
    squeakArgVec[squeakArgCnt++]= *skipArg();

# undef saveArg
# undef skipArg
}


int main(int argc, char **argv, char **envp)
{
  /* check the interpreter's size assumptions for basic data types */
  if (sizeof(int)    != 4) error("This C platform's integers are not 32 bits.");
  if (sizeof(double) != 8) error("This C platform's floats are not 64 bits.");
  if (sizeof(time_t) != 4) error("This C platform's time_t's are not 32 bits.");

  /* Make parameters global for access from pluggable primitives */
  argCnt= argc;
  argVec= argv;
  envVec= envp;

  /* Allocate arrays to store copies of pointers to command line
     arguments.  Used by getAttributeIntoLength(). */
  if ((vmArgVec= calloc(argc + 1, sizeof(char *))) == 0)
    outOfMemory();

  if ((squeakArgVec= calloc(argc + 1, sizeof(char *))) == 0)
    outOfMemory();

  signal(SIGSEGV, segv);

  /* initialisation */

# if defined(__alpha__)
  /* disable printing of unaligned access exceptions */
  {
    int buf[2]= { SSIN_UACPROC, UAC_NOPRINT };
    if (setsysinfo(SSI_NVPAIRS, buf, 1, 0, 0, 0) < 0)
      {
	perror("setsysinfo(UAC_NOPRINT)");
      }
  }
# endif

  progName= argv[0];
  ParseEnvironment();
  ParseArguments(argc, argv);
  SetUpTimers();

	AT_EXIT((void(*)(void))ioShutdownAllModules);

  sqFileInit();
  joystickInit();

# if defined(HAVE_TZSET)
  tzset();	/* should _not_ be necessary! */
# endif

  if (!realpath(argv[0], vmName))
    vmName[0]= 0; /* full VM name */

  /* read the image file and allocate memory for Squeak heap */
  {
    FILE *f= 0;

#   if defined(USE_INTERNAL_IMAGE) /* EXPERIMENTAL */
#ifdef USE_XMODEM
    if (using_xmodem)
      {
        xmodem_open();
        RecordFullPathForImageName("xmodem");
      }
    else
#endif
    if (openInternal())
      {
	RecordFullPathForImageName(shortImageName);
      }
    else
#   endif
      {
	if (0 == (f= fopen(shortImageName, "r")))
	  imageNotFound(shortImageName);
	else
	  RecordFullPathForImageName(shortImageName); /* full image path */
      }
    readImageFromFileHeapSize(f, initialHeapSize);
    sqImageFileClose(f);
  }


  openFbDisplay();

#ifdef TX39_TPANEL
  openTpanel();
#endif
#ifdef TX39_BUTTONS
  openButtons();
#endif

#if 0 /* TODO */
#ifndef HEADLESS
  /* open the Squeak window. */
  if (headless)
    forgetXDisplay();
  else
    openXDisplay();
#endif
#endif

#if defined(HAVE_LIBDL) && defined(J_ENABLED) && !defined(J_PROFILE)
  if (!noJitter)
    {
      extern void *loadExternalFunctionFromModule(char *, char *);
      /* first try to find an internal dynamic compiler... */
      void *comp= loadExternalFunctionFromModule("j_interpret", 0);
      /* ...and if that fails... */
      if (!comp)
	/* ...try to find an external one */
	comp= loadExternalFunctionFromModule("j_interpret", "SqueakCompiler");
      if (comp)
	{
	  ((void(*)(void))comp)();
	  return 0;
	}
    }
#endif /*HAVE_LIBDL*/

  /* run Squeak */
#ifndef J_PROFILE
  interpret();
#else
  j_interpret();
#endif

  return 0;
}

/*** Browser plugin stuff ***/



/* Return true if access to the given file is allowed.  Grant permission
   to access only those files in directories at or under the image directory. 
   Note: vmPath is actually the path to the image.
   NOTE ALSO: that this is broken.  Symbolic links will confuse it. -- ikp
 */
int plugInAllowAccessToFilePath(char *pathString, int pathStringLength)
{
  return true;
}

/* returns the size of the Squeak window */
int ioScreenSize(void)
{
  unsigned int w;
  unsigned int h;

  w = fb_vinfo.xres;
  h = fb_vinfo.yres;

  return (w << 16) | (h & 0xFFFF);  /* w is high 16 bits; h is low 16 bits */
}


time_t convertToSqueakTime(time_t unixTime)
{
#ifdef HAVE_TM_GMTOFF
  unixTime+= localtime(&unixTime)->tm_gmtoff;
#else
# ifdef HAVE_TIMEZONE
  unixTime+= ((daylight) * 60*60) - timezone;
# else
#  error: cannot determine timezone correction
# endif
#endif
  /* Squeak epoch is Jan 1, 1901.  Unix epoch is Jan 1, 1970: 17 leap years
     and 52 non-leap years later than Squeak. */
  return unixTime + ((52*365UL + 17*366UL) * 24*60*60UL);
}

/* returns the local wall clock time */
int ioSeconds(void)
{
  return convertToSqueakTime(time(0));
}


#if defined(USE_INTERNAL_IMAGE)

  /*-- EXPERIMENTAL -- read [and inflate] an internal image file --*/

  int sqImageFileClose(FILE *f)
  {
    int err= 0;
#ifdef USE_XMODEM
    if (using_xmodem) {
      xmodem_close();
      return 0;
    }
#endif
    if (f != 0)
      return fclose(f);
    assert(internalImage != 0);
    if (internalZStream != 0)
      {
	printf("decompressed %ld bytes\n", ztell(internalZStream));
	err= zclose(internalZStream);
      }
    internalImage= 0;
    internalImagePtr= 0;
    internalZStream= 0;
    return err;
  }

  int sqImageFilePosition(FILE *f)
  {
#ifdef USE_XMODEM
    if (using_xmodem) {
      return xmodem_position;
    }
#endif
    if (f != 0)
      return ftell(f);
    assert(internalImage != 0);
    if (internalZStream != 0)
      return ztell(internalZStream);
    return internalImagePtr - internalImage;
  }

  int sqImageFileRead(void *ptr, size_t sz, size_t count, FILE *f)
  {
#ifdef USE_XMODEM
    if (using_xmodem) {
      xmodem_read(ptr, sz * count);
      return sz * count;
    }
#endif
    if (f != 0)
      return fread(ptr, sz, count, f);
    assert(internalImage != 0);
    if (internalZStream != 0)
      return zread(ptr, sz, count, internalZStream);
    memcpy(ptr, (void *)internalImagePtr, sz * count);
    internalImagePtr+= sz*count;
    return sz*count;
  }

  int sqImageFileSeek(FILE *f, long pos)
  {
#ifdef USE_XMODEM
    if (using_xmodem) {
      return xmodem_seek(pos);
    }
#endif
    if (f != 0)
      return fseek(f, pos, SEEK_SET);
    assert(internalImage != 0);
    if (internalZStream != 0)
      return zseek(internalZStream, pos, SEEK_SET);
    internalImagePtr= internalImage + pos;
    return 0;
  }

  int openInternal(void)
  {
    unsigned char *internalImageEnd= 0;
    void *handle= dlopen(0, RTLD_NOW);
    if (handle == 0)
      {
	fprintf(stderr, "dlopen: %s\n", dlerror());
	exit(1);
      }
    /* non-zero means use in-memory file operations */
    internalImage= (unsigned char *)dlsym(handle, "__squeak_image_start");
    if (internalImage != 0)
      {
	internalImageEnd= (unsigned char *)dlsym(handle, "__squeak_image_end");
	internalImagePtr= internalImage;
	printf("reading internal image at 0x%08x + %d\n",
	       (unsigned)internalImage,
	       (int)(internalImageEnd - internalImage));
	strcpy(shortImageName, "internal.image");
      }
    else
      {
	internalImage= (unsigned char *)dlsym(handle, "__squeak_image_gz_start");
	if (internalImage == 0)
	  return 0;
	internalImageEnd= (unsigned char *)dlsym(handle, "__squeak_image_gz_end");
	{
	  char name[64], comment[64];
	  if (0 == gzstat(internalImage,
			  name, sizeof(name), comment, sizeof(comment)))
	    {
	      fprintf(stderr, "internal image: %s\n", z_error);
	      exit(1);
	    }
	  printf("decompressing %s at 0x%08x + %d\n",
		 ((*name == '\0') ? "internal image" : name),
		 (unsigned)internalImage,
		 (int)(internalImageEnd - internalImage));
	  if (*name == '\0')
	    strcpy(shortImageName, "internal.image");
	  else
	    strcpy(shortImageName, name);
	  if (*comment != '\0')
	    printf("%s\n", comment);
	}
	/* non-zero means inflate on-the-fly */
	internalZStream= gzopen(internalImage, internalImageEnd - internalImage);
	if (internalZStream == 0)
	  {
	    fprintf(stderr, "zopen: %s\n", z_error);
	    exit(1);
	  }
      }
    dlclose(handle);
    return internalImage != 0;
  }

#endif /* USE_INTERNAL_IMAGE */


void segv(int ignore)
{
  error("Segmentation fault");
}

void RecordFullPathForImageName(char *localImageName)
{
  struct stat s;
  /* get canonical path to image */
  if ((stat(localImageName, &s) == 0) || (realpath(localImageName, vmPath) == 0))
    {
      if (localImageName[0] == '/')
	{
	  strcpy(vmPath, localImageName);
	}
      else 
	{
	  getcwd(vmPath, sizeof(vmPath));
	  strcat(vmPath, "/");
	  strcat(vmPath, localImageName);
	}
    }
  strcpy(imageName, vmPath);
  /* truncate vmPath to dirname */
  {
    int i= 0;
    for (i= strlen(vmPath); i >= 0; i--)
      if ('/' == vmPath[i])
	{
	  vmPath[i+1]= '\0';
	  return;
	}
  }
  /* this might just work in an emergency... */
  strcpy(imageName, vmPath);
}

int imageNameSize(void)
{
  return strlen(imageName);
}

int imageNameGetLength(int sqImageNameIndex, int length)
{
  char *sqImageName= (char *)sqImageNameIndex;
  int count, i;

  count= strlen(imageName);
  count= (length < count) ? length : count;

  /* copy the file name into the Squeak string */
  for (i= 0; i < count; i++)
    sqImageName[i]= imageName[i];

  return count;
}

int imageNamePutLength(int sqImageNameIndex, int length)
{
  char *sqImageName= (char *)sqImageNameIndex;
  int count, i;

  count= (IMAGE_NAME_SIZE < length) ? IMAGE_NAME_SIZE : length;

  /* copy the file name into a null-terminated C string */
  for (i= 0; i < count; i++) imageName[i]= sqImageName[i];
  imageName[count]= 0;

#if 0
#ifndef HEADLESS
  /* update the window title */
  XStoreName(stDisplay, stWindow, imageName);
#endif
#endif

  return count;
}

#ifdef USE_ITIMER
void sigalrm(int signum)
{
  lowResMSecs+= LOW_RES_TICK_MSECS;
}
#endif


void SetUpTimers(void)
{
  /* set up the micro/millisecond clock */
  gettimeofday(&startUpTime, 0);
#ifdef USE_ITIMER
  /* set up the low-res (50th second) millisecond clock */
  /* WARNING: all system calls must check for EINTR!!! */
  {
    struct sigaction sa;
    sigset_t ss1, ss2;
    sigemptyset(&ss1);
    sigprocmask(SIG_BLOCK, &ss1, &ss2);
    sa.sa_handler= sigalrm;
    sa.sa_mask= ss2;
#ifdef SA_RESTART
    sa.sa_flags= SA_RESTART;
#else
    sa.sa_flags= 0;	/* assume that we have default BSD behaviour */
#endif
#ifdef __linux__
    sa.sa_restorer= 0;
#endif
    sigaction(SIGALRM, &sa, 0);
  }
  {
    struct itimerval iv;
    iv.it_interval.tv_sec= 0;
    iv.it_interval.tv_usec= LOW_RES_TICK_MSECS * 1000;
    iv.it_value.tv_sec= 0;
    iv.it_value.tv_usec= LOW_RES_TICK_MSECS;
    setitimer(ITIMER_REAL, &iv, 0);
  }
#endif
}




static int pend_dispBitsIndex, pend_width, pend_height, pend_depth, pend_affectedL, pend_affectedR, pend_affectedT, pend_affectedB;

int ioForceDisplayUpdate(void)
{
  /* pendingDisplay = 0; */

  /*    true_ioShowDisplay(pend_dispBitsIndex, pend_width, pend_height, pend_depth, pend_affectedL, pend_affectedR, pend_affectedT, pend_affectedB); */

  return 0;
}

#define swab32(x) \
        ((unsigned int)( \
                (((unsigned int)(x) & (unsigned int)0x000000ffUL) << 24) | \
                (((unsigned int)(x) & (unsigned int)0x0000ff00UL) <<  8) | \
                (((unsigned int)(x) & (unsigned int)0x00ff0000UL) >>  8) | \
                (((unsigned int)(x) & (unsigned int)0xff000000UL) >> 24) ))

static void ew_memcpy(int *dest, int *src, size_t n)
{
  int i;

  n = (n + 3) / 4;
  
  for (i = 0; i < n; i ++) {
    dest[i] = swab32(src[i]);
  }
}

#define bytesPerLine(width, depth)	((((width)*(depth) + 31) >> 5) << 2)
#define bytesPerLineRD(width, depth)	((((width)*(depth)) >> 5) << 2)

int ioShowDisplay(int dispBitsIndex, int width, int height, int depth,
		  int affectedL, int affectedR, int affectedT, int affectedB)
{
  /* TODO */

  if (affectedT == affectedB || affectedR == affectedL) return 0;

  if (affectedR > width) affectedR= width;
  if (affectedB > height) affectedB= height;

  if(affectedR < affectedL || affectedT > affectedB)
    return 1;

  /* fprintf(stderr, "%x w%d h%d d%d, %d %d %d %d\n", dispBitsIndex, width, height, depth, affectedL, affectedR, affectedT, affectedB); */

  ew_memcpy(fbmem, (char *)dispBitsIndex, (width*height*depth) / 8);
  /* copyReverseImageWords((int *)dispBitsIndex, (int *)fbmem,
                        depth, width, height,
                        affectedL, affectedT, affectedR, affectedB); */

  return 0;
}

#if 0
int ioShowDisplay(int dispBitsIndex, int width, int height, int depth,
		  int affectedL, int affectedR, int affectedT, int affectedB)
{
  pend_dispBitsIndex = dispBitsIndex;
  pend_width = width;
  pend_height = height;
  pend_depth = depth;

  true_ioShowDisplay(dispBitsIndex, width, height, depth, affectedL, affectedR, affectedT, affectedB);

  if (pendingDisplay) {
    /* just update bounds etc */
    if (affectedL < pend_affectedL) pend_affectedL = affectedL;
    if (affectedT < pend_affectedT) pend_affectedT = affectedT;
    if (affectedR > pend_affectedR) pend_affectedR = affectedR;
    if (affectedB < pend_affectedB) pend_affectedB = affectedB;
  } else {
    pendingDisplay = 1;
    pend_affectedL = affectedL;
    pend_affectedR = affectedR;
    pend_affectedT = affectedT;
    pend_affectedB = affectedB;
  }

}
#endif

int ioHasDisplayDepth(int i) 
{

  if (i == fb_vinfo.bits_per_pixel)
    return true;

#if 0
  switch (i) {
  case 1: return true;
  case 4:
    return false;
  case 8:
    /* return true; */
  case 16:
    return false;
  case 32:
    return false;
  }; 
#endif

  return false;
}


int ioSetDisplayMode(int width, int height, int depth, int fullscreenFlag)
{
  fprintf(stderr, "ioSetDisplayMode(%d, %d, %d, %d)\n",
	  width, height, depth, fullscreenFlag);
  return 0;
}



int ioExit(void)
{
  /* TODO */

  /*** Isn't this where... ***/

  exit(0);
}

int ioSetCursorWithMask(int cursorBitsIndex, int cursorMaskIndex,
			int offsetX, int offsetY)
{
  return 0;
}

int ioSetCursor(int cursorBitsIndex, int offsetX, int offsetY)
{
  return 0;
}

int ioSetFullScreen(int fullScreen)
{
  return 0;
}

int clipboardWriteFromAt(int count, int byteArrayIndex, int startIndex)
{
  return 0;
}

int clipboardReadIntoAt(int count, int byteArrayIndex, int startIndex)
{
  return 0;
}

int clipboardSize(void)
{
  return 0;
}

static char *getAttribute(int id)
{
  if (id < 0) {
    /* VM argument */
    if (-id  < vmArgCnt)
      return vmArgVec[-id ];
  } else switch (id) {
    case 0: 
      return vmArgVec[0];
    case 1:
      return imageName;
    case 1001:
      /* basic OS type: "unix", "win32", "mac", ... */
      return OS_TYPE;
    case 1002:
      /* specific OS type: "solaris2.5" on unix, "win95" on win32 */
      return VM_HOST_OS;
    case 1003:
      /* processor architecture: "68k", "x86", "PowerPC", ...  */
      return VM_HOST_CPU;
    case 1004:
      return  (char *)interpreterVersion;
    default:
      if ((id-2) < squeakArgCnt)
	return squeakArgVec[id-2];
  }
  return "";
}

int attributeSize(int id)
{
  return strlen(getAttribute(id));
}

int getAttributeIntoLength(int id, int byteArrayIndex, int length)
{
  if (length > 0)
    strncpy((char *)byteArrayIndex, getAttribute(id), length);
  return 0;
}

/*** Profiling ***/



int clearProfile(void) { return 0; }
int dumpProfile(void) { return 0; }
int startProfiling(void) { return 0; }
int stopProfiling(void) { return 0; }
